from abc import abstractmethod

from pymaya.request_classes.maya_base_request import MayaBaseRequest


class MayaSecurityBaseRequest(MayaBaseRequest):
    maya_api_base_url = "https://api.tase.co.il/api/"

    def __init__(self, *args, **kwargs):
        super(MayaSecurityBaseRequest, self).__init__(*args, **kwargs)

    @property
    @abstractmethod
    def end_point(self):
        pass

    @property
    @abstractmethod
    def method(self):
        pass



from abc import ABC, abstractmethod

from urllib.parse import urljoin

import requests


class MayaBaseRequest(ABC):
    def __init__(self, *args, **kwargs):
        self.request = requests.Request(self.method, **kwargs)
        self.request.headers["Cache-Control"] = "no-cache"
        self.request.headers["referer"] = "https://www.tase.co.il/"
        self.request.headers["User-Agent"] = "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; FSL 7.0.6.01001)"

    @property
    @abstractmethod
    def maya_api_base_url(self):
        pass

    @property
    @abstractmethod
    def end_point(self):
        pass

    @property
    @abstractmethod
    def method(self):
        pass

    def _get_url(self) -> str:
        return urljoin(self.maya_api_base_url, self.end_point)

    def prepare(self) -> requests.PreparedRequest:
        self.request.url = self._get_url()
        return self.request.prepare()



from logging import Logger
from typing import Dict
from datetime import date

from requests import Session

from pymaya.maya_funds import MayaFunds
from pymaya.maya_security import MayaSecurity
from pymaya.utils import Language


class Maya:
    def __init__(
        self,
        logger: Logger = None,
        num_of_attempts: int = 1,
        session: Session = Session(),
        verify: bool = True,
        cachesize: int = 128,
    ):

        self.maya_securities = MayaSecurity(
            logger=logger, num_of_attempts=num_of_attempts, session=session, verify=verify, cachesize=cachesize
        )

        self.maya_funds = MayaFunds(
            logger=logger, num_of_attempts=num_of_attempts, session=session, verify=verify, cachesize=cachesize
        )

        self.mapped_securities = {}
        self.map_securities()

    def get_all_securities(self, lang: Language = Language.ENGLISH):
        return self.maya_securities.get_all_securities(lang)

    def map_securities(self):
        all_securities = self.get_all_securities()
        for security in all_securities:
            if security.get("Id") in self.mapped_securities:
                self.mapped_securities[security.get("Id")].add(security.get("Type"))
            else:
                self.mapped_securities[security.get("Id")] = {(security.get("Type"))}

    def get_maya_class(self, security_id: str):
        if MayaFunds.TYPE in self.mapped_securities.get(security_id):
            return self.maya_funds
        else:
            return self.maya_securities

    def get_details(self, security_id: str, lang: Language = Language.ENGLISH):
        maya_class = self.get_maya_class(security_id)
        return maya_class.get_details(security_id, lang)

    def get_names(self, security_id: str):
        maya_class = self.get_maya_class(security_id)
        english_details = maya_class.get_details(security_id, lang=Language.ENGLISH)
        hebrew_details = maya_class.get_details(security_id, lang=Language.HEBREW)
        return maya_class.get_names(english_details, hebrew_details)

    def get_price_history_chunk(
        self, security_id: str, from_data: date, to_date: date, page: int, lang: Language = Language.ENGLISH
    ) -> Dict:
        maya_class = self.get_maya_class(security_id)
        return maya_class.get_price_history_chunk(
            security_id, from_data=from_data, to_date=to_date, page=page, lang=lang
        )

    def get_price_history(
        self,
        security_id: str,
        from_data: date,
        to_date: date = date.today(),
        page: int = 1,
        lang: Language = Language.ENGLISH,
    ):
        maya_class = self.get_maya_class(security_id)
        return maya_class.get_price_history(security_id, from_data=from_data, to_date=to_date, page=page, lang=lang)



class BadResponseException(Exception):
    def __init__(self, status_code, reason, url, method=None, *args) -> None:
        message = f"Bad response to URL: {url} in {method} method [status_code: {status_code}, reason: {reason}]"
        super(BadResponseException, self).__init__(message)



from pymaya.request_classes.maya_security_base_request import MayaSecurityBaseRequest
from pymaya.utils import Language


class MayaSecurityDataRequest(MayaSecurityBaseRequest):
    end_point = "company/securitydata"
    method = "GET"

    def __init__(self, security_id, lang: Language = Language.ENGLISH):
        super(MayaSecurityDataRequest, self).__init__(lang)
        self.request.params["securityId"] = security_id
        self.request.params["lang"] = int(lang)



from datetime import date

from pymaya.request_classes.maya_security_base_request import MayaSecurityBaseRequest
from pymaya.utils import Language


class MayaSecurityHistoricalRequest(MayaSecurityBaseRequest):
    end_point = "security/historyeod"
    method = "POST"

    def __init__(
        self,
        security_id: str,
        from_data: date,
        to_date: date,
        page: int = 1,
        p_type: int = 8,
        total_rec: int = 1,
        lang: Language = Language.ENGLISH,
    ):
        super(MayaSecurityBaseRequest, self).__init__()
        self.request.headers["Content-Type"] = "application/json"
        self.request.json = {
            "dFrom": str(from_data),
            "dTo": str(to_date),
            "oId": str(security_id),
            "pageNum": page,
            "pType": str(p_type),
            "TotalRec": total_rec,
            "lang": str(int(lang)),
        }



from typing import Optional
from datetime import date

from pymaya.request_classes.maya_fund_base_request import MayaFundBaseRequest
from pymaya.utils import Language


class MayaHistoricalRequest(MayaFundBaseRequest):
    end_point = "fund/history"
    method = "POST"

    def __init__(
        self,
        security_id: str,
        from_data: Optional[date],
        to_date: Optional[date],
        page: int = 1,
        period: int = 0,
        lang: Language = Language.ENGLISH,
    ):
        super(MayaHistoricalRequest, self).__init__(lang, data={})
        self.request.headers["Content-Type"] = "application/x-www-form-urlencoded"
        self.request.data["DateFrom"] = from_data.isoformat() if from_data else None
        self.request.data["DateTo"] = to_date.isoformat() if to_date else None
        self.request.data["FundId"] = security_id
        self.request.data["Page"] = page
        self.request.data["Period"] = period



from typing import List, Dict, Any
from datetime import date

from pymaya.maya_base import MayaBase
from pymaya.request_classes.maya_fund_details_request import MayaFundDetailsRequest
from pymaya.request_classes.maya_historical_request import MayaHistoricalRequest
from pymaya.utils import streamify, Language


class MayaFunds(MayaBase):
    TYPE = 4

    @staticmethod
    def get_names(english_details: Dict[str, Any], hebrew_details: Dict[str, Any]):
        return {
            "english_short_name": english_details.get("FundLongName", ""),
            "english_long_name": english_details.get("FundLongName", ""),
            "hebrew_short_name": hebrew_details.get("FundShortName", ""),
            "hebrew_long_name": hebrew_details.get("FundLongName", ""),
        }

    def get_details(self, security_id: str, lang: Language = Language.ENGLISH):
        return self._send_request(MayaFundDetailsRequest(security_id, lang))

    def get_price_history_chunk(
        self, security_id: str, from_data: date, to_date: date, page: int, lang: Language = Language.ENGLISH
    ) -> Dict:
        return self._send_request(MayaHistoricalRequest(security_id, from_data, to_date, page, lang=lang))

    @streamify
    def get_price_history(
        self,
        security_id: str,
        from_data: date,
        to_date: date = date.today(),
        page: int = 1,
        lang: Language = Language.ENGLISH,
    ) -> List[Dict]:
        data = self.get_price_history_chunk(security_id, from_data, to_date, page, lang)
        return data.get("Table", [])



from abc import abstractmethod

from pymaya.request_classes.maya_base_request import MayaBaseRequest
from pymaya.utils import Language


class MayaFundBaseRequest(MayaBaseRequest):
    maya_api_base_url = "https://mayaapi.tase.co.il/api/"

    def __init__(self, lang, *args, **kwargs):
        super(MayaFundBaseRequest, self).__init__(*args, **kwargs)
        self.request.headers["X-Maya-With"] = "allow"
        self.request.headers["Accept-Language"] = "en-US" if lang == Language.ENGLISH else "he-IL"

    @property
    @abstractmethod
    def end_point(self):
        pass

    @property
    @abstractmethod
    def method(self):
        pass



from pymaya.request_classes.maya_security_base_request import MayaSecurityBaseRequest
from pymaya.utils import Language


class MayaSecurityDetailsRequest(MayaSecurityBaseRequest):
    end_point = "security/majordata"
    method = "GET"

    def __init__(self, security_id, lang: Language = Language.ENGLISH):
        super(MayaSecurityDetailsRequest, self).__init__(lang)
        self.request.params["secId"] = security_id
        self.request.params["compId"] = security_id[:3]
        self.request.params["lang"] = int(lang)



from functools import partial
from typing import List, Dict, Any
from datetime import date

from cachetools import cachedmethod
from cachetools.keys import hashkey

from pymaya.maya_base import MayaBase
from pymaya.request_classes.maya_all_securities_request import MayaAllSecuritiesRequest
from pymaya.request_classes.maya_security_data_request import MayaSecurityDataRequest
from pymaya.request_classes.maya_security_request import MayaSecurityHistoricalRequest
from pymaya.utils import streamify, Language


class MayaSecurity(MayaBase):
    @cachedmethod(lambda self: self.cache, key=partial(hashkey, "all"))
    def get_all_securities(self, lang: Language = Language.ENGLISH):
        return self._send_request(MayaAllSecuritiesRequest(lang))

    @staticmethod
    def get_names(english_details: Dict[str, Any], hebrew_details: Dict[str, Any]):
        return {
            "english_short_name": english_details.get('Name'),
            "english_long_name": english_details.get('LongName') or english_details.get('SecurityLongName'),
            "hebrew_short_name": hebrew_details.get("Name"),
            "hebrew_long_name": hebrew_details.get("LongName") or hebrew_details.get('SecurityLongName'),
        }

    def get_details(self, security_id: str, lang: Language = Language.ENGLISH):
        return self._send_request(MayaSecurityDataRequest(security_id, lang=lang))

    def get_price_history_chunk(
        self, security_id: str, from_data: date, to_date: date, page: int, lang: Language = Language.ENGLISH
    ) -> Dict:
        return self._send_request(MayaSecurityHistoricalRequest(security_id, from_data, to_date, page, lang=lang))

    @streamify
    def get_price_history(
        self,
        security_id: str,
        from_data: date,
        to_date: date = date.today(),
        page: int = 1,
        lang: Language = Language.ENGLISH,
    ) -> List[Dict]:

        if not to_date:
            to_date = date.today()

        data = self.get_price_history_chunk(security_id, from_data, to_date, page, lang)
        return data.get("Items", [])



from pymaya.request_classes.maya_fund_base_request import MayaFundBaseRequest
from pymaya.utils import Language


class MayaFundDetailsRequest(MayaFundBaseRequest):
    end_point = "fund/details"
    method = "GET"

    def __init__(self, security_id, lang: Language = Language.ENGLISH):
        super(MayaFundDetailsRequest, self).__init__(lang)
        self.request.params["fundId"] = security_id



from functools import partial
from typing import Dict
from logging import Logger
import traceback
import time

import requests
from cachetools.keys import hashkey
from requests import Session
from cachetools import cachedmethod, LRUCache

from pymaya.request_classes.maya_base_request import MayaBaseRequest
from pymaya.utils import get_logger
from pymaya.exceptions import BadResponseException


class MayaBase:
    def __init__(
        self,
        logger: Logger = None,
        num_of_attempts: int = 1,
        session: Session = Session(),
        verify: bool = True,
        cachesize: int = 128,
    ):
        self.logger = get_logger(self.__class__.__name__) if logger is None else logger
        self.num_of_attempts = num_of_attempts
        self.session = session
        self.verify = verify
        self.cache = LRUCache(maxsize=cachesize)

    def get_response(self, maya_api_request: MayaBaseRequest):
        response = None

        for attempt in range(self.num_of_attempts):
            response = self.session.send(request=maya_api_request.prepare(), verify=self.verify)

            if response.status_code != requests.codes.ok:
                self.logger.info("response error attempt: {}".format(attempt + 1 / self.num_of_attempts))
                self.logger.info("status response code {}, reason {}".format(response.status_code, response.reason))
                self.logger.info("try again in 1 sec...")
                time.sleep(1)

            else:
                return response
        return response

    @cachedmethod(lambda self: self.cache, key=partial(hashkey, "send"))
    def _send_request(self, maya_api_request: MayaBaseRequest) -> Dict:
        try:
            response = self.get_response(maya_api_request=maya_api_request)

            if response.status_code != requests.codes.ok:
                self.logger.error("Error in API call [{}] - {}".format(response.status_code, response.reason))
                raise BadResponseException(
                    status_code=response.status_code,
                    reason=response.reason,
                    url=response.url,
                    method=response.request.method,
                )

            return response.json()

        except Exception as e:
            self.logger.error("Error {}, traceback: {}".format(e, traceback.format_exc()))
            raise



from pymaya.request_classes.maya_security_base_request import MayaSecurityBaseRequest
from pymaya.utils import Language


class MayaAllSecuritiesRequest(MayaSecurityBaseRequest):
    end_point = "content/searchentities"
    method = "GET"

    def __init__(self, lang: Language = Language.ENGLISH):
        super(MayaSecurityBaseRequest, self).__init__()
        self.request.headers["Content-Type"] = "application/json"
        self.request.params["lang"] = int(lang)



from typing import Callable, Iterable, Iterator, TypeVar
from enum import unique, IntEnum
from logging import Logger
import logging


T = TypeVar("T")


def streamify(batch_func: Callable[..., Iterable[T]]) -> Callable[..., Iterator[T]]:
    def streamify_wrapper(*args, **kwargs) -> Iterator[T]:
        kwargs["page"] = kwargs.get("page", 1)
        batch = batch_func(*args, **kwargs)
        while batch:
            yield from batch
            kwargs["page"] += 1
            batch = batch_func(*args, **kwargs)

    return streamify_wrapper


def get_logger(logger_name) -> Logger:
    logger = logging.getLogger(logger_name)
    stream_handler = logging.StreamHandler()
    formatter = logging.Formatter("[%(asctime)s %(levelname)s: %(message)s [in %(filename)s:%(lineno)d]")
    stream_handler.setLevel(logging.DEBUG)
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    logger.setLevel(logging.DEBUG)
    return logger


@unique
class Language(IntEnum):
    HEBREW = 0
    ENGLISH = 1



import setuptools

with open("README.md", "r") as fh:
    description = fh.read()

with open('requirements.txt') as f:
    install_reqs = f.readlines()
    reqs = [str(ir) for ir in install_reqs]

setuptools.setup(
    name="pymaya",
    version="0.5.0",
    author="Ben Sterenson",
    author_email="bensterenson@gmail.com",
    description="Python client to interact with maya.tase.co.il",
    install_requires=reqs,
    long_description=description,
    long_description_content_type="text/markdown",
    url="https://github.com/BenSterenson/pymaya",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    test_suite="tests",
)
